"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.mobileGetPermissions = exports.mobileChangePermissions = void 0;
const driver_1 = require("appium/driver");
const bluebird_1 = __importDefault(require("bluebird"));
const lodash_1 = __importDefault(require("lodash"));
const utils_1 = require("../utils");
const ALL_PERMISSIONS_MAGIC = 'all';
const PM_ACTION = Object.freeze({
    GRANT: 'grant',
    REVOKE: 'revoke',
});
const APPOPS_ACTION = Object.freeze({
    ALLOW: 'allow',
    DENY: 'deny',
    IGNORE: 'ignore',
    DEFAULT: 'default',
});
const PERMISSION_TARGET = Object.freeze({
    PM: 'pm',
    APPOPS: 'appops',
});
const PERMISSIONS_TYPE = Object.freeze({
    DENIED: 'denied',
    GRANTED: 'granted',
    REQUESTED: 'requested',
});
/**
 * @this {import('../driver').AndroidDriver}
 * @param {import('./types').ChangePermissionsOpts} opts
 * @returns {Promise<void>}
 */
async function mobileChangePermissions(opts) {
    const { permissions, appPackage = this.opts.appPackage, action = lodash_1.default.toLower(opts.target) === PERMISSION_TARGET.APPOPS
        ? APPOPS_ACTION.ALLOW
        : PM_ACTION.GRANT, target = PERMISSION_TARGET.PM, } = opts;
    if (lodash_1.default.isNil(permissions)) {
        throw new driver_1.errors.InvalidArgumentError(`'permissions' argument is required`);
    }
    if (lodash_1.default.isEmpty(permissions)) {
        throw new driver_1.errors.InvalidArgumentError(`'permissions' argument must not be empty`);
    }
    switch (lodash_1.default.toLower(target)) {
        case PERMISSION_TARGET.PM:
            return await changePermissionsViaPm.bind(this)(permissions, appPackage, lodash_1.default.toLower(action));
        case PERMISSION_TARGET.APPOPS:
            this.ensureFeatureEnabled(utils_1.ADB_SHELL_FEATURE);
            return await changePermissionsViaAppops.bind(this)(permissions, appPackage, lodash_1.default.toLower(action));
        default:
            throw new driver_1.errors.InvalidArgumentError(`'target' argument must be one of: ${lodash_1.default.values(PERMISSION_TARGET)}`);
    }
}
exports.mobileChangePermissions = mobileChangePermissions;
/**
 * @this {import('../driver').AndroidDriver}
 * @param {import('./types').GetPermissionsOpts} [opts={}]
 * @returns {Promise<string[]>}
 */
async function mobileGetPermissions(opts = {}) {
    const { type = PERMISSIONS_TYPE.REQUESTED, appPackage = this.opts.appPackage } = opts;
    /**
     * @type {(pkg: string) => Promise<string[]>}
     */
    let actionFunc;
    switch (lodash_1.default.toLower(type)) {
        case PERMISSIONS_TYPE.REQUESTED:
            actionFunc = (pkg) => this.adb.getReqPermissions(pkg);
            break;
        case PERMISSIONS_TYPE.GRANTED:
            actionFunc = (pkg) => this.adb.getGrantedPermissions(pkg);
            break;
        case PERMISSIONS_TYPE.DENIED:
            actionFunc = (pkg) => this.adb.getDeniedPermissions(pkg);
            break;
        default:
            throw new driver_1.errors.InvalidArgumentError(`Unknown permissions type '${type}'. ` +
                `Only ${JSON.stringify(lodash_1.default.values(PERMISSIONS_TYPE))} types are supported`);
    }
    return await actionFunc(/** @type {string} */ (appPackage));
}
exports.mobileGetPermissions = mobileGetPermissions;
// #region Internal helpers
/**
 * @this {AndroidDriver}
 * @param {string|string[]} permissions
 * @param {string} appPackage
 * @param {import('type-fest').ValueOf<PM_ACTION>} action
 */
async function changePermissionsViaPm(permissions, appPackage, action) {
    if (!lodash_1.default.values(PM_ACTION).includes(action)) {
        throw new driver_1.errors.InvalidArgumentError(`Unknown action '${action}'. ` +
            `Only ${JSON.stringify(lodash_1.default.values(PM_ACTION))} actions are supported`);
    }
    let affectedPermissions = lodash_1.default.isArray(permissions) ? permissions : [permissions];
    if (lodash_1.default.isString(permissions) && lodash_1.default.toLower(permissions) === ALL_PERMISSIONS_MAGIC) {
        const dumpsys = await this.adb.shell(['dumpsys', 'package', appPackage]);
        const grantedPermissions = await this.adb.getGrantedPermissions(appPackage, dumpsys);
        if (action === PM_ACTION.GRANT) {
            const reqPermissons = await this.adb.getReqPermissions(appPackage, dumpsys);
            affectedPermissions = lodash_1.default.difference(reqPermissons, grantedPermissions);
        }
        else {
            affectedPermissions = grantedPermissions;
        }
        if (lodash_1.default.isEmpty(affectedPermissions)) {
            this.log.info(`'${appPackage}' contains no permissions to ${action}`);
            return;
        }
    }
    if (action === PM_ACTION.GRANT) {
        await this.adb.grantPermissions(appPackage, affectedPermissions);
    }
    else {
        await bluebird_1.default.all(affectedPermissions.map((name) => this.adb.revokePermission(appPackage, name)));
    }
}
/**
 * @this {AndroidDriver}
 * @param {string|string[]} permissions
 * @param {string} appPackage
 * @param {import('type-fest').ValueOf<APPOPS_ACTION>} action
 */
async function changePermissionsViaAppops(permissions, appPackage, action) {
    if (!lodash_1.default.values(APPOPS_ACTION).includes(action)) {
        throw new driver_1.errors.InvalidArgumentError(`Unknown action '${action}'. ` +
            `Only ${JSON.stringify(lodash_1.default.values(APPOPS_ACTION))} actions are supported`);
    }
    if (lodash_1.default.isString(permissions) && lodash_1.default.toLower(permissions) === ALL_PERMISSIONS_MAGIC) {
        throw new driver_1.errors.InvalidArgumentError(`'${ALL_PERMISSIONS_MAGIC}' permission is only supported for ` +
            `'${PERMISSION_TARGET.PM}' target. ` +
            `Check AppOpsManager.java from Android platform sources to get the full list of supported AppOps permissions`);
    }
    const promises = (lodash_1.default.isArray(permissions) ? permissions : [permissions]).map((permission) => this.adb.shell(['appops', 'set', appPackage, permission, action]));
    await bluebird_1.default.all(promises);
}
// #endregion
/**
 * @typedef {import('appium-adb').ADB} ADB
 * @typedef {import('../driver').AndroidDriver} AndroidDriver
 */
//# sourceMappingURL=permissions.js.map